# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
from datetime import timedelta

from zope.interface import Interface, Attribute
from zope.i18nmessageid import MessageFactory
from zope.schema import ASCIILine, Password, Timedelta, Bool, Choice, \
     Timedelta, TextLine, Int, Text, \
     Tuple

from dm.saml2.util import SAML_CLASSES, normalize_class, \
     ATTRNAME_FORMATS, normalize_attrname_format, \
     XSCHEMA_BASE_TYPES, \
     NAMEID_FORMATS

from util import vocab_from_urns

_ = MessageFactory('dm_zope_saml2')

##############################################################################
# role interfaces

class IRole(Interface):
  """An SAML role."""

class IIdpssoRole(IRole):
  """An identity provider."""

class ISpssoRole(IRole):
  """A service provider."""

  def add_attribute_metadata(descriptor):
    """add information about the attribute consuming services to metadata *descriptor*."""

class IApRole(IRole):
  """An attribute provider."""

  def add_attribute_metadata(descriptor):
    """add information about the provided attributes to metadata *descriptor*."""

class IAuthnRole(IRole):
  """An authentication authority."""

class IPdpRole(IRole):
  """An policy decision point."""



##############################################################################
# schemata


class IItemSchema(Interface):
  """the schema describing a base item (elementary Zope object).

  It also has an id, manage outside.
  """
  title = TextLine(
    title=_(u"title_title", u"Title"),
    required=False,
    default=u"",
    )


class ISamlAuthoritySchema(IItemSchema):
  """Parameters configuring an Saml authority (aka entity)."""
  entity_id = ASCIILine(
    title=_(u'entity_id_title', u'Entity id'),
    description=
    _(u'entity_id_description',
      u"""The id identifying this entity/authority.""",),
    required=True,
    )
    
  certificate = ASCIILine(
    title=_(u'certificate_title', u'Certificate'),
    description=
    _(u'certificate_description',
      u"""`clienthome` relative or absolute path to the (DER) file containing the certificate corresponding to 'Private key'.""",),
    required=False,
    )

    
  future_certificate = ASCIILine(
    title=_(u'future_certificate_title', u'Future certificate'),
    description=
    _(u'future_certificate_description',
      u"""`clienthome` relative or absolute path to the (DER) file containing the certificate you plan to use in the near future.
      As other SAML2 authorities rely on your metadata published certificates
      to verify your signatures,
      you cannot simply change your privat/public key pair and publish
      a new certificate: until the other parties had updated their
      metadata for you, they would be unable to verify your signature
      signed with your new private key. This field allows you to publish
      in advance the certificate for a new private key you plan to
      use in the near future. If you ensure a sufficient delay,
      they can be prepared for the key change.
      """,),
    required=False,
    )

  private_key = ASCIILine(
    title=_(u'private_key_title', u'Private key'),
    description=
    _(u'private_key_description',
      u"""`clienthome` relative or absolute path to the (PEM) file containing the private key used for signing/encryption."""
      ),
    required=False,
    )

  private_key_password = Password(
    title=_(u'private_key_password_title', u'Private key password'),
    description=
    _(u'private_key_password_description',
      u"""Password used to encrypt the private key"""),
    required=False
    )

  base_url = ASCIILine(
    title=_(u'base_url_title', u'Base url'),
    description=
    _(u'base_url_description',
      u"""A Zope system is often used via different urls (e.g. direct versus indirectly via a Web server).
      The urls of internal objects change accordingly. The authority generates
      and distributes metadata involving url. These must remain
      reliably and not change inadvertantly. Therefore, the base url is
      not derived automatically from the (varying) urls but specified by this attribute."""),
    required=True,
    )

  metadata_validity = Timedelta(
    title=_(u'metadata_validity_title', u'Metadata validity'),
    description=
    _(u'metadata_validity_description',
      u"""Validity period of generated metadata."""),
    required=True,
    default=timedelta(1), # 1 day
    )


class IBaseAttributeSchema(Interface):
  """base schema describing an attribute.

  Note: it also has an `id` managed outside. The `id` is also used
  as attribute name with which the member is accessed to obtain the
  attribute value.
  """
  title = TextLine(
    title=_(u"attribute_name_title", u"External attribute name"),
    description=_(
      u"attribute_name_description",
      u"The name used to identify the attribute across SAML entities."
      ),
    required=True,
    )

  format = Choice(
    title=_(u"attribute_format_title", u"Name format"),
    description=_(
      u"attribute_format_description",
      u"The name format used to identify this attribute"
      ),
    required=True,
    vocabulary=vocab_from_urns(ATTRNAME_FORMATS),
    default=normalize_attrname_format("unspecified"),
    )

  description = Text(
    title=_(u"attribute_description_title", u"Description"),
    description=_(
      u"attribute_description_description",
      u"Informational (not used internally) description of the attribute",
      ),
    required = False,
    )


class IProvidedAttributeSchema(IBaseAttributeSchema):
  """schema for a provided attribute."""
  type = Choice(
    title=_(u"attribute_type_title", u"Attribute type"),
    description=_(
      u"attribute_type_description",
      u"The XML-schema type of this attribute. It must be an base type. Tuples and lists of base the base type are automatically handled."
      ),
    values=XSCHEMA_BASE_TYPES,
    required=True,
    default=u"string",
    )

  evaluator = ASCIILine(
    title=_(u"attribute_evaluator_title", u"Evaluator"),
    description=_(
      u"attribute_evaluator_description",
      u"Name of a method or view to evaluate/determine the attribute's value. "
      u"It is called with *member*, "
      u"*attr* (providing `IProvidedAttributeSchema`) and *eid* (the entity "
      u"requesting the attribute)."
      ),
    required=False,
    )


class IRequestedAttributeSchema(IBaseAttributeSchema):
  """schema for a requested attribute."""
  required = Bool(
    title=_(u"attribute_required_title", u"Required?"),
    description=_(
      u"attribute_required_description",
      u"Is this attribute required?",
      ),
    required=True,
    default=False,
    )

  is_sequence = Bool(
    title=_(u"attribute_is_sequence_title", u"Sequence?"),
    description=_(
      u"attribute_is_sequence_description",
      u"Has this attribute a sequence as value?"
      ),
    required=True,
    default=False,
    )


class IAttributeConsumingServiceSchema(IItemSchema):
  # override "title" to make it required
  title = TextLine(
    title=_(u"title_title", u"Title"),
    required=True,
    default=u"",
    )

  index = Int(
    title=_(u"attribute_consuming_service_index_title", u"Index"),
    description=_(
      u"attribute_consuming_service_index_description",
      u"Unique integer index identifying this service.",
      ),
    required=True,
    default=0,
    )

  is_default = Bool(
    title=_(u"attribute_consuming_service_is_default_title", u"Default?"),
    description=_(
      u"attribute_consuming_service_is_default_description",
      u"Is this the default service. Only one service must be the default!.",
      ),
    required=True,
    default=False,
    )

  language = ASCIILine(
    title=_(u"attribute_consuming_service_language_title", u"Language"),
    description=_(
      u"attribute_consuming_service_language_description",
      u"The language of the service description (e.g. `en`, `de`, `fr` ...)."
      ),
    required=True,
    default="en",
    )

  description = Text(
    title=_(u"service_description_title", u"Description"),
    description=_(
      u"service_description_description",
      u"Information describing the service.",
      ),
    required = False,
    )

  extends = ASCIILine(
    title=_(u"attribute_consuming_service_extends_title", u"Extends"),
    description=_(
      u"attribute_consuming_service_extends_description",
      u"blank separated sequence of service ids this service extends (not yet supported)"),
    required=False,
    default="",
    )
  

class ISamlRoleSchema(IItemSchema):
  """Common configuration data for SAML role providers.

  See `ISimpleIdpSchema` for the dependence between a role and an
  authority/entity.
  """

  # not yet: assume a global registration
##  authority_designator = ASCIILine(
##    title=_(u'authority_designator_title', u'Authority'),
##    description=
##    _(u'authority_designator_description',
##      u"""path or utility name identifying the SAML authority (= entity) this role belongs to."""),
##    required=False,
##    )


class ISimpleIdpSchema(ISamlRoleSchema):
  """Schema describing the configuration data of a simple identity provider.

  While one can view an identity provider as an SAML2 authority,
  we prefer to associate the authority with an SAML2 entity and
  view the identity provider as its role. This implies a bidirectional
  link between the role and the entity.
  """

  authn_context_class = Choice(
    title=_(u"authn_context_class_title", "Authentication context class"),
    description=_(
      u"auth_context_class_description",
      u"The authentication context class implemented by this identity provider."
      u" The value is likely either `Password` (HTTP username/password authentication) or `PasswordProtectedTransport` (HTTPS username/password authentication)."),
    vocabulary=vocab_from_urns(SAML_CLASSES),
    required=True,
    default=normalize_class("Password"),
    )

  # later
##  idp_discovery_domain = ASCIILine(
##    title=_(u"idp_discovery_domain_title", u"IDP discovery domain"),
##    description=
##    _(u"idp_discovery_domain_description",
##      u"Domain used to determine the applicable identity providers"),
##    required=False,
##    )

##  authn_confirmation_required = Bool(
##    title=_(u"authn_confirmation_required_title", u"authentication confirmation required"),
##    description=
##    _(u"authn_confirmation_required_description",
##      u"whether a confirmation is required before we redirect to a new service provider"),
##    default=False,
##    required=True,
##    )

##  authn_confirmation_validity = Timedelta(
##    title=_(u"authn_confirmation_validity_title", u"authentication confirmation validity"),
##    description=
##    _(u"authn_confirmation_validity_description",
##      u"time period an authentication confirmation for a service provider is valid -- defaults to the current session."),
##    required=False,
##    )

##  max_session_validity = Timedelta(
##    title=_(u"max_session_validity_title", u"maximal session validity"),
##    description=
##    _(u"max_session_validity_description",
##      u"By default, the authentication session is coupled to the browser session; if specified this parameter limits the lifetime of the authentication session. Implementation specifics may limit the duration as well."),
##    required=False,
##    )
    


class ISimpleSpSchema(ISamlRoleSchema):

  default_authn_context_class = Choice(
    title=_(u"default_authn_context_class_title", "Default authentication context class"),
    description=_(
      u"default_auth_context_class_description",
      u"The default authentication context class required by this service provider."
      u" The value is likely either `Password` (HTTP username/password authentication) or `PasswordProtectedTransport` (HTTPS username/password authentication)."),
    vocabulary=vocab_from_urns(SAML_CLASSES),
    required=False,
    )

  cookie_path = ASCIILine(
    title=_(u"cookie_path_title", u"Cookie path"),
    description=_(u"cookie_path_description",
                  u"Path used for all cookies"),
    required=True,
    default="/"
    )

  cookie_domain = ASCIILine(
    title=_(u"cookie_domain_title", u"Cookie domain"),
    description=_(u"cookie_domain_description",
                  u"Domain used for all cookies"),
    required=False,
    )

  session_cookie_name = ASCIILine(
    title=_(u"session_cookie_name_title", u"Session cookie name"),
    description=_(u"session_cookie_name_description",
                  u"The name of the cookie representing the authentication information"),
    required=True,
    default="_saml2_session",
    )

  attribute_cookie_name = ASCIILine(
    title=_(u"attribute_cookie_name_title", u"Attribute cookie name"),
    description=_(u"attribute_cookie_name_description",
                  u"The name of the cookie holding SAML2 attribute information"),
    required=False,
    default="_saml2_attributes",
    )

  encrypt_cookies = Bool(  
    title=_(u"encrypt_cookies_title", u"should cookies be encrypted"),
    description=_(u"encrypt_cookies_description",
                  u"controls whether cookie values are encrypted"),
    required=False,
    default=False,
    )

  nameid_formats = Tuple(
    title=_(u"nameid_formats_title", u"Nameid formats"),
    description=_(u"name_id_formats_description",
                  u"Specifies which nameid formats are acceptable. If left empty, any nameid format is acceptable."
                  ),
    value_type=Choice(
      vocabulary=vocab_from_urns(NAMEID_FORMATS),
      ),
    required=False,
    default=(),
    )

  allow_create = Bool(
    title=_(u"allow_create_title", u"Allow identifier creation"),
    description=_(
      u"allow_create_description",
      u"Allows the identity provider to create/associate a new identifier for the authenticated user."),
    required=False,
    default=True,
    )

  # always
##  wants_assertions_signed = Bool(
##    title=_(u"wants_assertions_signed_title", u"Wants assertions signed?"),
##    description=_(
##      u"wants_assertions_signed_description",
##      u"Indicated whether assertions must be signed?"
##      ),
##    required=False,
##    default=False,
##    )



class ISimpleSpssoPluginSchema(IItemSchema):
  """schema describing a PAS Plugin working together with a SimpleSpsso."""
  idp_cookie_name = ASCIILine(
    title=_(u"idp_cookie_name_title", u"Identity provider cookie name"),
    description=_(u"idp_cookie_name_description",
                  u"The name of the cookie used to remember the user's identity provider"),
    required=False,
    default="idp_id",
    )

  idp_cookie_path = ASCIILine(
    title=_(u"idp_cookie_path_title", u"Identity provider cookie path"),
    description=_(u"idp_cookie_path_description",
                  u"The path of the cookie used to remember the user's identity provider; defaults to the portal"),
    required=False,
    )

  idp_cookie_domain = ASCIILine(
    title=_(u"idp_cookie_domain_title", u"Identity provider cookie domain"),
    description=_(u"idp_cookie_domain_description",
                  u"The domain of the cookie used to remember the user's identity provider"),
    required=False,
    )

  idp_cookie_lifetime = Timedelta(
    title=_(u"idp_cookie_lifetime_title", u"Identity provider cookie lifetime"),
    description=_(u"idp_cookie_lifetime_description",
                  u"The lifetime of the cookie used to remember the user's identity provider. If not specified, this is a session cookie. Example values are `1d` (1 day), `3600s` (1 hour)."),
    required=False,
    default=timedelta(360), # 1 year
    )

  default_idp = ASCIILine(
    title=_(u"default_idp_title", u"Default identity provider"),
    description=_(u"default_idp_description",
                  u"The idp used as default in idp selection"),
    required=False,
    )

  failure_view = ASCIILine(
    title=_(u"failure_view_title", u"Failure view"),
    description=_(u"failure_view_description",
                  u"The view presenting SAML problems."),
    required=True,
    default="@@saml_failure",
    )

  select_idp_view = ASCIILine(
    title=_(u"select_idp_view_title", u"Select idp view"),
    description=_(u"select_idp_view_description",
                  u"The view used to select an identity provider."),
    required=True,
    default="@@saml_select_idp",
    )



class ISimpleApSchema(ISamlRoleSchema):
  """Schema describing the configuration data of a simple attribute provider."""

  # not yet
##  attr_confirmation_required = Bool(
##    title=_(u"attr_confirmation_required_title", u"attribute confirmation required"),
##    description=
##    _(u"attr_confirmation_required_description",
##      u"whether a confirmation is required before we deliver (new) attribute information to a service provider -- defaults to the current session."),
##    default=False,
##    required=True,
##    )

##  attr_confirmation_validity = Timedelta(
##    title=_(u"attr_confirmation_validity_title", u"authentication confirmation validity"),
##    description=
##    _(u"attr_confirmation_validity_description",
##      u"time period an attribute confirmation for a service provider is valid"),
##    required=False,
##    )



class ISamlAuthority(ISamlAuthoritySchema):
  """SAML authority (= entity) interface.

  An SAML authority provides common services for a set of roles.
  It manages the metadata describing itself and that of a set of other SAML
  authorities.
  """
  def register_role_implementor(implementor):
    """register *implementor* as an instance implementing one or more roles for this authority."""

  def unregister_role_implementor(implementor):
    """unregister *implementor*."""

  def export_own_metadata():
    """export this authorities metadata."""

  def export_all_metadata():
    """export the metadata for all authorities known by this authority."""

  def add_entity(entity):
    """add *entity* to the list of managed entities."""

  def list_entities():
    """the list of known entities.

    Note that the entities must not be modified.
    """

  def del_entity(self, eid):
    """delete entity *eid*."""


class ISimpleIdpsso(ISimpleIdpSchema, IIdpssoRole):
  """a simple identity provider."""


class ISimpleSpsso(ISimpleSpSchema, ISpssoRole):
  """a simple service provider."""

  def get_authentication_session(request):
    """return a mapping describing the authentication session (or `None`).

    Return `None`, when there is no authentication session or it is no longer
    valid.

    Otherwise, return a mapping with at least the `user_id` key.
    The `user_id` identifies the user in the form *idp*::*id*
    where *idp* specifies the identity provider which has performed
    the authentication and *id* it the user id it has associated with
    the user for this service provider.

    The mapping may contain additional keys such as
    `authn_context_class` (the authentication context class),
    `authn_time` (when did the authentication took place)
    and `valid_until` (up to which time is the authentication valid).
    """

  def invalidate_authentication_session(request):
    """invalidate all information associated with this user."""

  def get_attributes(request):
    """return a mapping describing the users SAML2 attributes (or `None`).

    Return `None`, when there is no authentication session or it is no longer
    valid.

    Otherwise, return a mapping with attribute values provided by
    the SAML2 identity or attribute provider. The attributes are
    identified by their friendly names.
    """

  def authenticate(idp, ok, fail, authn_context_class=None, passive=False, force=False, acs_index=None, REQUEST=None):
    """request authentication from *idp* and redirect to *ok* or *fail*.

    *acs_index* specifies the index of an attribute consuming service.
    """

  def clear_keys():
    """clear all keys used for signing/verification/encryption/decryption.

    As a consequence, all authentication sessions and attributes
    become invalid.
    """

  def new_key():
    """create a new key used for signing/encryption.

    This causes a new key to be generated - it is used for signing/encryption
    of new values. Some number of older keys are still available for
    verification/decryption. `new_key` may cause the oldest key to be no
    longer used for anything.

    The default number of verification/decryption keys is 5.
    """


class ISimpleAttributeProvider(ISimpleApSchema, IApRole):
  """a simple attribute provider."""



# auxiliary definition to work around Zope's restructuring
try: from zope.lifecycleevent.interfaces import IObjectMovedEvent
except ImportError:
  # old Zope, old location
  from zope.app.container.interfaces import IObjectMovedEvent


###############################################################################
## HttpTransport
class IHttpTransport(Interface):
  """Transport used for SOAP binding.

  Attention: to satisfy the SAML2 security requirements, this
  transport must use an HTTPS client certificate and verify the HTTPS
  server certificate for https connections.
  """

  def open(url, data, headers):
    """post *data* to *url* with *headers* as (some) request headers.

    return a file like object (similar to `urllib.urlopen`).
    """


###############################################################################
## Encryption
class IEncryption(Interface):
  """encryption utility.

  A corresponding utility must be registered for an `SPSSO` role to
  encrypt its cookie values.
  """
  def encrypt(key, value):
    """*value* encrypted by *key*."""

  def decrypt(key, value):
    """*value* decrypted by *key*."""


#############################################################################
## NameID format support
class INameidFormatSupport(Interface):
  """handle name id format extensions."""
  supported = Attribute(u"the sequence of supported name id formats.")

  def make_id(member, sp, ok_create, nid):
    """generate an id for *member*.

    return `None`, if the generation fails, otherwise the id content.

    *sp* is the service provider id. Some formats require ids dependent
    on this id (such as `persistent`).

    *ok_create* is a boolean specifying whether new ids can be created.

    *nid* is a `NameID` instance, partially filled. Its `Format`
    attribute specifies the required nameid format. Some formats
    require the its `SPNameQualifier` be set (e.g. `persistent`).
    """


class IEntity(Interface):
  """An entity, a standard SAML authority can work with."""

  id = Attribute(u"the entity id")

  title = Attribute(u"the entities title. "
                    u"Used as human readable entity representation. "
                    u"Defaults to the entity id."
                    )

  def getId():
    """valid Zope id, determined from `id` by escaping."""



"""Entity presentation."""
from zope.interface import Interface
from zope.schema import Bool
from zope.formlib.form import Fields, action

from dm.zope.schema.z2.template import form_template
from dm.zope.schema.z2.form import ZmiMixin, PageForm

from dm.zope.saml2.interfaces import _

class IUpdateSchema(Interface):
  clear_existing_metadata = Bool(
    title=_(u"clear_existing_metadata", u"Clear existing metadata"),
    required=False,
    )

class Update(ZmiMixin, PageForm):
  label = _(u"entity_metadata_update", u"Metadata update")
  form_fields = Fields(IUpdateSchema)

  @action(_(u"update_metadata", u"Update metadata"))
  def update_(self, action, data):
    e = self.context
    e.aq_inner.aq_parent.metadata_by_id(e.id).fetch_metadata(data["clear_existing_metadata"])
    self.status = _(u"metadata_updated", u"Metadata updated")

  







# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Identity provider SSO.

We assume to live inside a (Plone) portal and can rely on its login
infrastructure (probably with an adapted login form).

We assume the existance of a local `ISAMLAuthority` utility. It may be inside
the portal itself but more likely it lives above the portal to be usable
by several portals.

For special services, such as `SingleLogout`, different
authentication strenght, limited authentication
session lifetime, correct authentication time, cookie authentication,
more portal infrastructure will need to be adapted
(especially, the `Session` plugin).
We plan to provide adequate base implementations in the future.
"""




# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""The SPSSO role."""
from datetime import datetime
try: from hashlib import sha256 as digest_module
except ImportError: import md5 as digest_module
import hmac
from cPickle import loads, dumps
from zlib import compress, decompress

from zope.interface import implements
from zope.component import getUtility

from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from ZTUtils import make_query
from persistent.list import PersistentList

from dm.saml2.util import normalize_class, compare_classes, pyxb_to_datetime, \
     normalize_attrname_format, xs_convert_from_xml
from dm.zope.schema.schema import SchemaConfigured

from dm.zope.saml2.interfaces import ISimpleSpsso, \
     INameidFormatSupport, \
     IEncryption
from dm.zope.saml2.permission import manage_saml
from dm.zope.saml2.sso import Sso
from dm.zope.saml2.role import Target
from dm.zope.saml2.attribute import \
     HomogenousContainer, AttributeConsumingService


class SimpleSpsso(HomogenousContainer, Sso):
  """Zope 2 implementation of a simple SAML2 Spsso."""

  implements(ISimpleSpsso)

  SC_SCHEMAS = (ISimpleSpsso,)
  CONTENT_TYPE = AttributeConsumingService

  security = ClassSecurityInfo()

  security.declareObjectProtected(manage_saml)
  security.declarePublic("authenticate")

  # as we only support http-post and this requires signing
  wants_assertions_signed = True

  # newly introduced
  nameid_formats = ()
  allow_create = True

  
  def __init__(self, **kw):
    SchemaConfigured.__init__(self, **kw)
    Sso.__init__(self)
    self.__keys = PersistentList()
    self.new_key()

  def authenticate(self, idp, ok, fail, authn_context_class=None, passive=False, force=False, acs_index=None, REQUEST=None):
    """authenticate via *idp*."""
    r = REQUEST or self.REQUEST; R = r.response
    if authn_context_class is None:
      authn_context_class = self.default_authn_context_class
    if authn_context_class is not None:
      authn_context_class = normalize_class(authn_context_class)
    if not force:
      # see whether we have a valid authentication satisfying the requirements
      session = self.get_authentication_session(r)
      if session:
        comparison = (
          authn_context_class is None and -1
          or compare_classes(authn_context_class,
                             session["authn_context_class"]
                             )
          )
        if comparison is not None and comparison <= 0:
          return R.redirect(ok)
    # must authenticate
    from dm.saml2.pyxb.protocol import AuthnRequest, RequestedAuthnContext, \
         NameIDPolicy
    from dm.saml2.pyxb.assertion import AuthnContextClassRef
    req = AuthnRequest(ForceAuthn=force, IsPassive=passive)
    if authn_context_class is not None:
      req.RequestedAuthnContext = RequestedAuthnContext(
        AuthnContextClassRef(authn_context_class)
        )
    if acs_index is not None:
      req.AttributeConsumingServiceIndex = acs_index
    self.customize_authn_request(req)
    relay_state = self.store((req.ID, ok, fail))
    nip = NameIDPolicy(AllowCreate=self.allow_create)
    nifs = INameidFormatSupport(self).supported
    if len(nifs) == 1: nip.Format = nifs[0]
    req.NameIDPolicy = nip
    return self.deliver(
      Target(eid=idp, role="idpsso", endpoint="SingleSignOnService",
             sign_msg_attr="WantAuthnRequestsSigned",
             ),
      None, req, relay_state
      )

  # optionally overridden by derived classes to customize the request
  def customize_authn_request(self, req): pass

  def _process_AuthnStatement(self, subject, s):
    info = dict(
      name_qualifier=subject.NameQualifier,
      nameid_format=subject.Format,
      sp_name_qualifier=subject.SPNameQualifier,
      nameid=subject.value(),
      authn_time=pyxb_to_datetime(s.AuthnInstant),
      valid_until=pyxb_to_datetime(s.SessionNotOnOrAfter),
      session_id=s.SessionIndex,
      authn_context_class=s.AuthnContext.AuthnContextClassRef,
      )
    info["user_id"] = self.format_user_id(info)
    self._set_cookie(self.session_cookie_name, info)

  def _process_AttributeStatement(self, subject, s):
    # build map of known attributes -- we might want to cache this information
    attrs = {}
    for acs in self.objectValues():
      for att in acs.objectValues():
        attrs[(att.format, att.title)] = att
    # now process the attributes
    info = {}
    for att in s.Attribute:
      d = attrs.get(
        (att.NameFormat or normalize_attrname_format("unspecified"),
         att.Name
         ))
      if d is None: continue # do not know this attribute
      info[d.getId()] = xs_convert_from_xml(att.AttributeValue, d.is_sequence)
    self._set_cookie(self.attribute_cookie_name, info)



  def get_authentication_session(self, request):
    info = self._get_cookie(request, self.session_cookie_name)
    if info is None: return
    if info["valid_until"] is not None:
      if datetime.utcnow() >= info["valid_until"]: return # no longer valid
    return info


  def invalidate_authentication_session(self, request):
    expire = request.response.expireCookie
    params = self._cookie_params()
    expire(self.session_cookie_name, **params)
    expire(self.attribute_cookie_name, **params)


  def get_attributes(self, request):
    # only called when there is an auth session
    #auths = self.get_authentication_session()
    #if auths is None: return
    return self._get_cookie(request, self.attribute_cookie_name)


  def format_user_id(self, info):
    """format user id from dict *info*.

    Almost surely, the `nameid` key will be used.
    To ensure uniqueness, you might need the `name_qualifier`.
    `nameid_format` may allow specific decisions about
    the user id format.

    The primary purpose of this method is customization.
    Be warned however, that changing the format will invalidate
    existing user references.
    """
    return "%(name_qualifier)s::%(nameid)s" % info


  def _cookie_params(self):
    params = dict(path=self.cookie_path, httpOnly=True)
    if self.cookie_domain: params["domain"] = self.cookie_domain
    return params

  def _set_cookie(self, name, info):
    self.REQUEST.response.setCookie(name, self._encode(info),
                                    **self._cookie_params()
                                    )

  def _get_cookie(self, req, name):
    return self._decode(req.cookies.get(name))

  def _encode(self, info):
    """encode *info* for use as integrity protected cookie value."""
    s = self._serialize(info)
    key = self.__keys[0]
    if self.encrypt_cookies:
      value = getUtility(IEncryption).encrypt(key, s)
    else:
      value = hmac.new(key, s, digest_module).digest() + s
    return value.encode("base64").replace("\n", "").replace(" ", "")

  def _decode(self, v):
    """decode cookie value *v* into an info object (or `None`)."""
    if not v: return
    v = v.decode("base64")
    deserialize = self._deserialize
    if self.encrypt_cookies:
      decrypt = getUtility(IEncryption).decrypt
      for k in self.__keys:
        # try to decrypt and deserialize
        try: return deserialize(decrypt(key, v))
        except:
          # probably a wrong key
          pass
      return None # unable to understand the cookie
    else:
      dl = hmac.new("", "", digest_module).digest_size
      digest = v[:dl]; v = v[dl:]
      for k in self.__keys:
        if digest == hmac.new(k, v, digest_module).digest():
          return deserialize(v)
      return None # unable to authenticate

  def _serialize(self, obj):
    return compress(dumps(obj, -1), 9)

  def _deserialize(self, s):
    return loads(decompress(s))

  def add_attribute_metadata(self, descriptor):
    from dm.saml2.pyxb.metadata import AttributeConsumingService, \
         RequestedAttribute, ServiceName, ServiceDescription
    from dm.saml2.pyxb.assertion import Attribute
    # check uniqueness of "index" -- should probably happen earlier
    seen_indexes = set()
    for acs in self.objectValues():
      ads = acs.objectValues()
      if not ads: continue
      index = acs.index
      if index in seen_indexes:
        raise ValueError("duplicate acs index %d for sp %s"
                         % (index, descriptor.getId())
                         )
      seen_indexes.add(index)
      acs_md = AttributeConsumingService(
        ServiceName(acs.title, lang=acs.language),
        index=index, isDefault=acs.is_default,
        )
      if acs.description:
        acs_md.ServiceDescription.append(
          ServiceDescription(asc.description, lang=acs.language)
          )
      for ad in ads:
        acs_md.RequestedAttribute.append(
          RequestedAttribute(
            NameFormat=ad.format,
            Name=ad.title,
            FriendlyName=ad.id,
            isRequired=ad.required,
            )
          )
      descriptor.AttributeConsumingService.append(acs_md)

      

  ## key handling
  KEY_LENGTH = 32
  KEY_NUMBER = 5

  def _create_key(self):
    kl = self.KEY_LENGTH
    try:
      from os import urandom
      bytes = urandom(kl)
    except ImportError: # no "urandom"
      from random import randrange
      bytes = [chr(randrange(256)) for i in range(kl)]
    return "".join(bytes)

  def clear_keys(self):
    del self.__keys[:]

  def new_key(self):
    self.__keys.insert(0, self._create_key())
    del self.__keys[self.KEY_NUMBER:]

InitializeClass(SimpleSpsso)


class StandaloneSimpleSpsso(SimpleSpsso):
  """a `SimpleSpsso` that works with separate plugins.

  Integration is via a (local) utility registration.
  """
  meta_type = "Saml simple spsso (standalone)"




def initialize(context):
  from dm.zope.schema.z2.constructor import add_form_factory, SchemaConfiguredZmiAddForm

  context.registerClass(
    StandaloneSimpleSpsso,
    constructors=(add_form_factory(StandaloneSimpleSpsso, form_class=SchemaConfiguredZmiAddForm),),
    permission=manage_saml,
    )


### automatic [un]registration on add/move/delete
def move_handler(o, e):
  # import will probably need to move for newer Zope
  from zope.app.component.hooks import getSite
  site = getSite()
  if site is None:
      raise ValueError("need a persistent active site")
  sm = site.getSiteManager()
  if e.oldParent:
    # unregister
    sm.unregisterUtility(o, provided=ISimpleSpsso)
  if e.newParent:
    # register
    # ensure this is a persistent registry
    if not hasattr(sm, "_p_jar"):
      raise ValueError("need a persistent active site")
    sm.registerUtility(o, provided=ISimpleSpsso)


class NameidFormatSupport(object):
  """SPSSO name id format support."""
  implements(INameidFormatSupport)

  def __init__(self, context): self.context = context

  def make_id(*args):
    raise NotImplementedError # we only read name ids

  @property
  def supported(self): return self.context.nameid_formats



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Common base implementation for SSO support."""

from role import Role

class Sso(Role):
  """mixin class to provide common SSO services.
  """



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""The Idpsso implementation."""
from datetime import datetime, timedelta

from zope.interface import implements

from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from ZTUtils import make_query

from Products.CMFCore.utils import getToolByName

from dm.saml2.uuid import uuid
from dm.saml2.util import compare_classes, normalize_nameid_format
from dm.zope.schema.schema import SchemaConfigured
from dm.saml2.pyxb.assertion import AssertionType

from dm.zope.saml2.interfaces import ISimpleIdpsso, IApRole, \
     INameidFormatSupport
from dm.zope.saml2.permission import manage_saml
from dm.zope.saml2.sso import Sso
from dm.zope.saml2.role import Target
from dm.zope.saml2.attribute import SimpleAttributeProvider


# might want to implement as a tool
class SimpleIdpsso(SimpleItem, SchemaConfigured, Sso):
  """Zope2 implementation for a simple SAML Idpsso.

  This is Zope 2, thus we access the request via acquisition.
  Note, that this prevents access via a local utility.
  """
  meta_type = "Saml simple idpsso"

  # how long should authentication responses be valid
  BROWSER_SSO_VALIDITY = timedelta(minutes=5)

  implements(ISimpleIdpsso)

  SC_SCHEMAS = (ISimpleIdpsso,)

  security = ClassSecurityInfo()

  security.declareObjectProtected(manage_saml)
  security.declarePublic("idpsso_logged_in")

  manage_options = (
    {"label" : "View", "action" : "@@view"},
    {"label" : "Edit", "action" : "@@edit"},
    ) + SimpleItem.manage_options


# Standard Plone values -- change, if you need something special
  login_form = "acl_users/credentials_cookie_auth/require_login"
  came_from = "came_from"

  def handle_AuthnRequest(self, req, binding, relay_state):
    mt = getToolByName(self, "portal_membership")
    member = not mt.isAnonymousUser() and mt.getAuthenticatedMember() or None
    if req.IsPassive and not member:
      return self._failedAuthnRequest(req, relay_state, "NoPassive")
    # check for unsupported things -- may want to make this more flexible
    unsupported = []
    if req.Conditions: unsupported.append("Conditions")
    if req.Scoping: unsupported.append("Scoping")
    if unsupported:
      return self._failAuthnRequest(req, relay_state, "RequestUnsupported", ", ".join(unsupported))
    # check requested authn context
    ac = req.RequestedAuthnContext
    if ac:
      if ac.AuthnContextDeclRef:
        # we only support `AuthnContextClassRef`.
        return self._failAuthnRequest(req, relay_state, "RequestUnsupported", "AuthnContextDeclRef")
      if not self._supported_authn_context_class(ac.AuthnContextClassRef):  
        return self._failAuthRequest(req, relay_state, "NoAuthnContext")
    # check name id policy
    nip = req.NameIDPolicy
    if nip and not self._supported_nameid_policy(nip):
      return self._failAuthnRequest(req, relay_state, "InvalidNameIDPolicy")
    if req.ForceAuthn or member is None:
      if req.ForceAuthn: self.acl_users.resetCredentials() # force logout
      # prepare authentication
      r = self.REQUEST; s = r["SESSION"]
      skey = uuid()
      s[skey] = (req, relay_state)
      # the view logic gives us an extremely funny "URL" -- explicitely, use
      #  our URL
      # use portal url as base for "login_form"
      r.response.redirect("%s/%s?%s" % (
        getToolByName(self, "portal_url")(), self.login_form,
        make_query({self.came_from : "%s/idpsso_logged_in?%s" % (
          self.absolute_url(), make_query(skey=skey)
          )}),
        ))
    else: return self._okAuthnRequest(req, relay_state, member)


  def _make_authn_response_target(self, req=None):
    if req is None: params = {}
    else:
      params = dict(
        endpoint_index = req.AssertionConsumerServiceIndex,
        binding = req.ProtocolBinding,
        url = req.AssertionConsumerServiceURL,
        )
    return Target(role="spsso",
                  endpoint="AssertionConsumerService",
                  sign_ass_attr="WantAssertionsSigned",
                  **params
                  )

  def _okAuthnRequest(self, req, relay_state, member):
    target = self._make_authn_response_target(req)
    authn_ass = self._make_authn_assertion(target, req, member)
    if not isinstance(authn_ass, AssertionType):
      # this is in fact error information (a string for the moment)
      return self._failAuthnRequest(req, relay_state, authn_ass)
    # we require effectively that `req` is not `None`. This implies
    #  we do not support Idp initiated login.
    if IApRole.providedBy(self):
      attr_st = self._make_attribute_statement(
        target,
        req,
        authn_ass.Subject,
        member,
        req.AttributeConsumingServiceIndex,
        )
      if attr_st is not None: # attributes have been called for
        from dm.saml2.pyxb.assertion import AttributeStatementType
        if not isinstance(attr_st, AttributeStatementType):
          # this is in fact error information (a string for the moment)
          return self._failAuthnRequest(req, relay_state, attr_st)
        authn_ass.AttributeStatement.append(attr_st)
    return self.deliver_success(target, req, (authn_ass,), relay_state)

  def _make_authn_assertion(self, target, req, member):
    from dm.saml2.pyxb.assertion import SubjectType, AuthnStatement, \
         AuthnContext, AuthnContextClassRef
    ass = self.make_assertion()
    subject = self.subject_from_member(member, target, req)
    if not isinstance(subject, SubjectType): return subject # an error
    # add `SubjectConfirmation` and `AudienceRestriction`
    #   as required by the `Web Browser SSO Profile`
    # We might want to put this into a mixin class as
    #   it is helpful for any use of the HTTP-Post binding.
    from dm.saml2.pyxb.assertion import \
         SubjectConfirmation, SubjectConfirmationData, \
         Conditions, AudienceRestriction, Audience
    target.resolve(self._get_authority(), req)
    subject.SubjectConfirmation.append(     
      SubjectConfirmation(
        SubjectConfirmationData(
          NotOnOrAfter=datetime.utcnow() + self.BROWSER_SSO_VALIDITY,
          Recipient=target.url,
          InResponseTo=req.ID,
          ),
        Method="urn:oasis:names:tc:SAML:2.0:cm:bearer"
        ))
    cs = ass.Conditions = Conditions()
    cs.AudienceRestriction.append(AudienceRestriction(Audience(target.eid))
      )
    ass.Subject = subject
    # do we need `Conditions`?
    ass.AuthnStatement.append(
      AuthnStatement(
        AuthnContext(
          AuthnContextClassRef(self.authn_context_class)
          ),
        AuthnInstant=ass.IssueInstant, # we cheet here
        )
      )
    return ass


  def _failAuthnRequest(self, req, relay_state, code, msg=None):
    target = self._make_authn_response_target(req)
    return self.deliver_failure(target, req, code, relay_state, msg) 
                    
    
  def idpsso_logged_in(self, skey):
    """return here after a successful login."""
    r = self.REQUEST; s = r["SESSION"]
    req, relay_state = s[skey]
    member = getToolByName(self, "portal_membership").getAuthenticatedMember()
    return self._okAuthnRequest(req, relay_state, member)

  def _supported_authn_context_class(self, ac):
    # Note: "ac" is a sequence of context class references
    # In principle, we should take into account as well the `Comparison`
    #   However, the processing rules are stupid. Thus, we always assume
    #   "minimum".
    # We support a single context class.
    my_ac = self.authn_context_class
    for cr in ac:
      comparison = compare_classes(cr, my_ac)
      if comparison is not None and comparison <= 0: return True
    return False

  def _supported_nameid_policy(self, nip):
    # we only support "unspecified"
    return not nip.Format or nip.Format in INameidFormatSupport(self).supported

# no longer necessary
#InitializeClass(SimpleIdpsso)


class SimpleIdpssoAp(SimpleAttributeProvider, SimpleIdpsso):
  """a simple idpsso combined with a simple attribute provider."""
  meta_type = "Saml simple idpsso with integrated attribute provider"

  SC_SCHEMAS = SimpleIdpsso.SC_SCHEMAS + SimpleAttributeProvider.SC_SCHEMAS


def initialize(context):
  from dm.zope.schema.z2.constructor import add_form_factory, SchemaConfiguredZmiAddForm

  for cls in SimpleIdpsso, SimpleIdpssoAp:
    context.registerClass(
      cls,
      constructors=(add_form_factory(cls, form_class=SchemaConfiguredZmiAddForm),),
      permission=manage_saml,
      )




# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Role views."""
from logging import getLogger

from transaction import abort

from Products.Five.browser import BrowserView

from dm.saml2.binding.soap import http_request as soap_request
from dm.saml2.binding.httpredirect import decode as redirect_decode
from dm.saml2.binding.httppost import decode as post_decode
from dm.saml2.pyxb.protocol import CreateFromDocument

from dm.zope.saml2.role import logging_enabled

logger = getLogger(__name__)


class RoleView(BrowserView):
  """Role view.
  
  It implements the various bindings (the incoming part)
  and delegates request processing to its `IRole` context.
  """

  def soap(self):
    """the soap binding."""
    r = self.request
    soap = r.get("SOAPXML") or r["BODY"]
    result, headers, status = soap_request(soap, self._process)
    if status != 200: abort()
    # set response headers
    R = r.response
    R.setHeader("Content-Type", "text/xml; charset=utf-8") # this is SOAP 1.1
    for hn, hv in headers.items(): R.setHeader(hn, hv)
    # should we process the result?
    return result

  def redirect(self):
    """the HTTP redirect binding."""
    r = self.request
    msg, relay_state = redirect_decode("%s?%s" % (r["URL"], r["QUERY_STRING"]))
    return self._process(msg, binding="redirect", relay_state=relay_state)

  def post(self):
    msg, relay_state = post_decode(self.request.form)
    return self._process(msg, binding="post", relay_state=relay_state)

  # we do not yet support "artifact"


  def _process(self, saml, **kw):
    __traceback_info__ = saml
    # Determine our url: due to a bug in the view machinery, the
    #  last component can be repeated -- remove, if necessary
    r = self.request
    url1, url2 = r["URL1"], r["URL"]
    suff = url2[len(url1):]
    url = not url1.endswith(suff) and url2 or url1
    if logging_enabled:
      logger.info(
        "incoming saml message at %s via binding %s with relay state %s:\n%s"
        % (url,
           kw.get("bindung"),
           kw.get("relay_state"),
           saml
           )
        )
    saml = CreateFromDocument(saml)
    # If there are signatures they have been verified by now
    #  but we have not yet checked that there are indeed signatures
    # verify destination; set, if not available
    if saml.Destination and saml.Destination != url:
      raise SamlError("destination mismatch")
    saml.Destination = url # necessary for further checks
    return getattr(
      self.context, "handle_" + saml._element().name().localName()
      )(saml, **kw)



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""SPSSO role.

The architecture for this role consists of the role itself and a
cooperating `PlonePAS` plugin. In the simplest (and usual) case,
both fall together.
However, for special purposes (several sites share a common service provider),
role and plugin can be separate objects where the role is used by
several plugins.

The role does not have a user interface of itself. It operates as a
purely internal service. Therefore, it can be instantiated outside of
a portal context. The role is responsible for all SAML2 handling.
It communicates with the plugin indirectly via cookies. Therefore,
role and plugin must share those cookies. The plugin treats those
cookies as opaque values and uses a role method to access its content.
As a consequence, authentication sessions and attribute values
are shared among plugins. It is impossible that the same user
has different identities or different attributes in sites using
the same SPSSO role.

User ids have the form *NameQualifier*::*Id*, where *NameQualifier*
identifies the authenticating entity and *Id* specifies the id assigned
by it for the user in our usage. This means, that some information
from the `NameID` used in SAML2 is lost.

The plugin implements the `IExtractionPlugin`,
`IAuthenticationPlugin`, `ICredentialResetPlugin` and `IChallengePlugin`.
If you use the standard Plone `ISessionPlugin`, some SAML2 features
(such as global logout, lifetime limitation of the authentication session)
may not work (as the session plugin continues to authenticate the user);
therefore, you may consider its deactivation.
If you want to allow local authentication beside SAML2 authentication,
you need your own challenge plugin whose login form support idp selection
and local login.

Sensible cookie values can optionaly be encrypted. In this case,
an `IEncryption` utility must be registered. This package contains
an example implementation based on the `mcrypt` extension, but
does not register it by default.
If not encrypted, sensible cookie values are signed
(and thereby protected against manipulation).

The role operates together with an SAML authority, accessed as
an `ISamlAuthority` utility. A corresponding registration must
be available in the context of each plugin and the role.
"""

def initialize(context):
  from spsso import initialize; initialize(context)
  from plugin import initialize; initialize(context)



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Attribute handling.

Attributes occur in SAML2 in idp operations, sp operations and ap operations.
idps and aps provide attributes while sps consume them.
sps organize their attributes into services identified by an integer index.

We essentially have a class describing a single attribute (with a
variant describing a requested attribute), a class managing attributes
(and its variant used for service descriptions) and an attribute
provider class."""
from logging import getLogger

from zope.interface import implements

from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem
from OFS.Folder import Folder

from dm.zope.schema.schema import SchemaConfigured
from dm.zope.schema.z2.constructor import \
     add_form_factory, SchemaConfiguredZmiAddForm

from dm.saml2.util import xs_convert_to_xml

from interfaces import IProvidedAttributeSchema, IRequestedAttributeSchema, \
     IItemSchema, IAttributeConsumingServiceSchema, \
     ISimpleAttributeProvider
from permission import manage_saml
from role import Role

logger = getLogger(__name__)


class BaseAttribute(SchemaConfigured, SimpleItem):
  security = ClassSecurityInfo()
  security.declareObjectProtected(manage_saml)

  manage_options = (
    {"label" : "View", "action" : "@@view"},
    {"label" : "Edit", "action" : "@@edit"},
    )


class RequestedAttribute(BaseAttribute):
  meta_type = "Saml requested attribute"

  implements(IRequestedAttributeSchema)

  SC_SCHEMAS = (IRequestedAttributeSchema,)


class ProvidedAttribute(BaseAttribute):
  meta_type = "Saml provided attribute"

  implements(IProvidedAttributeSchema)

  SC_SCHEMAS = (IProvidedAttributeSchema,)


class HomogenousContainer(SchemaConfigured, Folder):
  """Abstract base class for the implementation of homogenous containers.

  Note: `HomogenousContainer` must be sufficiently high up in the `mro`
  to ensure that its `__class_init__` is called. The default
  `__class_init__` (set on `Persistent` by `App.PersistentExtras`)
  does not use cooperative super calling.
  """

  implements(IItemSchema)
  SC_SCHEMAS = (IItemSchema,)

  # to be overridden by derived classes
  CONTENT_TYPE = None

  security=ClassSecurityInfo()
  security.declareObjectProtected(manage_saml)

  manage_options = (
    Folder.manage_options[0],
    {"label" : "View", "action" : "@@view"},
    {"label" : "Edit", "action" : "@@edit"},
    ) + Folder.manage_options[3:]
    

  def all_meta_types(self):
    att = self.CONTENT_TYPE
    if att is None: return ()
    atn = att.__name__
    return dict(
      name=att.meta_type,
      action="add_" + atn,
      permission=manage_saml,
      ),

  @classmethod
  def __class_init__(cls):
    att = cls.CONTENT_TYPE
    if att is not None:
      add_form = add_form_factory(att,
                                  form_class=SchemaConfiguredZmiAddForm
                                  )
      afn = add_form.__name__
      setattr(cls, afn, add_form)
      security = getattr(cls, "security", None)
      if security is None: security = cls.security = ClassSecurityInfo()
      security.declareProtected(manage_saml, afn)
      scls = super(HomogenousContainer, cls)
    elif cls.__name__ == "HomogenousContainer":
      # cannot use the name "HomogenourContainer" here as it is not yet bound
      scls = super(cls, cls)
    else: raise SystemError("class %s has not set `CONTENT_TYPE`" % str(cls))
    scls.__class_init__.im_func(cls)


class AttributeContainer(HomogenousContainer):
  """simple container for attribute descriptions."""
  meta_type = "Saml attribute container"

  CONTENT_TYPE = ProvidedAttribute


class AttributeConsumingService(HomogenousContainer):
  meta_type = "Saml attribute consuming service"

  implements(IAttributeConsumingServiceSchema)
  SC_SCHEMAS = (IAttributeConsumingServiceSchema,)

  CONTENT_TYPE = RequestedAttribute


class SimpleAttributeProvider(AttributeContainer, Role):
  """Attribute provider.

  For the moment, we only provide attribute support for the `Idpsso`
  but do not handle `AttributeRequest` of our own. This will come later.
  """
  meta_type = "Saml attribute provider"

  implements(ISimpleAttributeProvider)
  SC_SCHEMAS = ISimpleAttributeProvider,

  def _make_attribute_statement(self, target, req, subject, member, index):
    eid = target.eid or req.Issuer.value()
    auth = self._get_authority()
    md = auth.metadata_by_id(eid).get_recent_metadata()
    for sp in md.SPSSODescriptor:
      for acs in sp.AttributeConsumingService:
        if index is None and acs.isDefault or index == acs.index: break
      else: acs = None
      if acs is not None: break
    else: acs = None
    if acs is None:
      if index is None: return # nothing to do
      logger.error("could not locate acs %d for %s" % (index, eid))
      return "ResourceNotRecognized"
    # catalog our attribute -- should probably be cached
    attrs = {}
    for att in self.objectValues(): attrs[(att.format, att.title)] = att
    # determine the attributes we are ready to provide
    from dm.saml2.pyxb.assertion import AttributeStatement, Attribute, AttributeValue
    av = []
    for ra in acs.RequestedAttribute:
      ran = (ra.NameFormat or normalize_attrname_format("unspecified"), ra.Name)
      d = attrs.get(ran)
      if d is None:
        logger.error("attribute %s requested by %s not found" % (ran, eid))
        continue
      evaluator = d.evaluator
      if evaluator is None: v = member.getProperty(d.getId(), None)
      else: v = self.unrestrictedTraverse(evaluator)(member, d, eid)
      # Plone stupidly converts unicode properties to `str`
      if isinstance(v, str) and d.type == "string":
        # convert back to unicode
        from Products.PlonePAS.utils import getCharset
        v = unicode(v, getCharset(self))
      # potentially, more encodings are necessary
      xv = xs_convert_to_xml(d.type, v, AttributeValue)
      if not isinstance(xv, list): xv = xv,
      aas = dict(
        NameFormat=d.format,
        Name=d.title,
        FriendlyName=ra.FriendlyName or d.getId(),
        )
      att = Attribute(*xv, **aas)
      av.append(att)
    if not av: return
    return AttributeStatement(*av)

  def add_attribute_metadata(self, descriptor):
    from dm.saml2.pyxb.assertion import Attribute
    for ad in self.objectValues():
      descriptor.Attribute.append(
        Attribute(NameFormat=ad.format,
                  Name=ad.title,
                  FriendlyName=ad.id,
                  )
        )



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""PAS plugins for SAML2 based authentication (Spsso)."""
from datetime import datetime

from zope.interface import implements

from AccessControl import ClassSecurityInfo
from ZTUtils import make_query
from Globals import InitializeClass

from Products.CMFCore.utils import getToolByName
from Products.PluggableAuthService.interfaces.plugins \
        import IExtractionPlugin, IAuthenticationPlugin, \
                ICredentialsResetPlugin, IChallengePlugin, \
                IPropertiesPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.PluggableAuthService import \
     registerMultiPlugin

from dm.zope.schema.schema import SchemaConfigured

from dm.zope.saml2.permission import manage_saml
from dm.zope.saml2.util import datetime_rfc822
from dm.zope.saml2.interfaces import ISimpleSpssoPluginSchema, \
     ISimpleSpsso

from spsso import SimpleSpsso


class DetachedSimpleSpssoPlugin(BasePlugin, SchemaConfigured):
  """Spsso plugin working with a separate `Spsso`."""

  meta_type = "Saml detached simple spsso plugin (with standalone spsso)."

  security = ClassSecurityInfo()
  security.declareObjectProtected(manage_saml)
  security.declarePublic("login")

  implements(ISimpleSpssoPluginSchema)

  manage_options = (
    {"label" : "View", "action" : "@@view"},
    {"label" : "Edit", "action" : "@@edit"},
    ) + BasePlugin.manage_options

  SC_SCHEMAS = (ISimpleSpssoPluginSchema,)

  def extractCredentials(self, request):
    return dict(saml=True) # may not play well with caching

  def authenticateCredentials(self, credentials):
    if not credentials.get("saml"): return
    spsso = self.get_spsso()
    info = spsso.get_authentication_session(self.REQUEST)
    if info is not None:
      uid = info["user_id"]
      return (uid, uid)

  def resetCredentials(self, request, response):
    self.get_spsso().invalidate_authentication_session(request)


  def challenge(self, request, response):
    url = self.absolute_url()
    came_from = "%s?%s" % (
      request["ACTUAL_URL"], request.get("QUERY_STRING","")
      )
    idp = self.determine_idp()
    if idp:
      # we know the idp -- perform authentication
      #  Note: we must do a redirect to `authenticate` as it
      #  requires a successful transaction commit.
      #  The current transaction, however, will be aborted
      spsso = self.get_spsso()
      url = "%s/authenticate?%s" % (
        spsso.absolute_url(),
        make_query(idp=idp, ok=came_from, fail=url + "/" + self.failure_view)
        )
    else:
      url = "%s/%s?%s" % (
          url, self.select_idp_view, make_query(came_from=came_from)
        )
    # we hope the protocol chooser will ensure higher up user folders
    #  will still be able to get their chance
    #  otherwise, we will need to override "response._unauthorized"
    #  rather than performing the redirect here ourselves.
    response.redirect(url, lock=True)
    return True

  def authn(self, idp, came_from, fail=None, authn_context_class=None, passive=False, force=False, ap_index=None):
    idp = idp or self.determine_idp()
    if not idp: raise ValueError("Identity provider should not be determined")
    return self.get_spsso().authenticate(
      idp,
      came_from,
      fail or self.absolute_url() + "/" + self.failure_view,
      authn_context_class,
      passive,
      force,
      ap_index,
      self.REQUEST,
      )

  def login(self):
    """explicit login request."""
    r = self.REQUEST; purl = getToolByName(self, "portal_url")()
    came_from = r.get("HTTP_REFERER") or purl
    idp = self.determine_idp()
    if idp:
      return self.authn(idp, came_from,
                        self.absolute_url() + "/" + self.failure_view
                        )
    else:
      r.response.redirect("%s/%s?%s" % (
        self.absolute_url(),
        self.select_idp_view,
        make_query(came_from=came_from)
        ))

  def getPropertiesForUser(self, user, request=None):
    info = self.get_spsso().get_attributes(self.REQUEST) or {}
    # the stupid Plone is unable to handle unicode properties
    #  must encode them
    from Products.PlonePAS.utils import getCharset
    charset = getCharset(self)
    for k,v in info.items():
      if isinstance(v, unicode): info[k] = v.encode(charset)
      elif v and isinstance(v, (tuple, list)) and isinstance(v[0], unicode):
        info[k] = [c.encode(charset) for c in v]
    # more conversion might need to become necessary
    return info

  def determine_idp(self):
    r = self.REQUEST
    idp_cn = self.idp_cookie_name
    idp = idp_cn and r.get(idp_cn)
    if not idp:
      idps = self.list_idps()
      if len(idps) == 1: idp = idps[0]
    if idp and idp_cn not in r.cookies:
      self.set_idp_cookie(idp)
    return idp or None

  def list_idps(self):
    auth = self.get_spsso()._get_authority()
    idps = []
    for e in auth.list_entities():
      md = auth.metadata_by_id(e.id).get_recent_metadata()
      if md.IDPSSODescriptor: idps.append(e.id)
    return idps

  def get_spsso(self):
    return ISimpleSpsso(self)

  def get_entity(self, eid):
    """return the entity specified by entity id *eid*."""
    return self.get_spsso()._get_authority().get_entity(eid)

  def set_idp_cookie(self, idp):
    cn = self.idp_cookie_name
    if not cn: return
    cp = self.idp_cookie_path
    if not cp:
      # default is the portal path
      cp = "/" + getToolByName(self, "portal_url").getPortalObject().absolute_url(True)
    cparams = dict(path=cp, httpOnly=True)
    if self.idp_cookie_lifetime:
      expires = datetime.now() + self.idp_cookie_lifetime
      # convert to rfc822
      cparams["expires"] = datetime_rfc822(expires)
    if self.idp_cookie_domain:
      cparams["domain"] = self.idp_cookie_domain
    self.REQUEST.response.setCookie(cn, idp, **cparams)


InitializeClass(DetachedSimpleSpssoPlugin)

classImplements(DetachedSimpleSpssoPlugin,
                IExtractionPlugin, IAuthenticationPlugin,
                ICredentialsResetPlugin, IChallengePlugin,
                IPropertiesPlugin,
                )
    
  
class IntegratedSimpleSpssoPlugin(SimpleSpsso, DetachedSimpleSpssoPlugin):
  meta_type = "Saml integrated simple spsso plugin (integrated spsso)."

  SC_SCHEMAS = (ISimpleSpssoPluginSchema, ISimpleSpsso,)

  manage_options = (
    SimpleSpsso.manage_options[0],
    ) + DetachedSimpleSpssoPlugin.manage_options




def initialize(context):
  from dm.zope.schema.z2.constructor import add_form_factory, SchemaConfiguredZmiAddForm

  registerMultiPlugin(DetachedSimpleSpssoPlugin.meta_type)
  registerMultiPlugin(IntegratedSimpleSpssoPlugin.meta_type)

  context.registerClass(
    DetachedSimpleSpssoPlugin,
    constructors=(add_form_factory(DetachedSimpleSpssoPlugin, form_class=SchemaConfiguredZmiAddForm),),
    permission=manage_saml,
    visibility=None,
    )

  context.registerClass(
    IntegratedSimpleSpssoPlugin,
    constructors=(add_form_factory(IntegratedSimpleSpssoPlugin, form_class=SchemaConfiguredZmiAddForm),),
    permission=manage_saml,
    visibility=None,
    )
  

def default_plugin_spsso_adapter(plugin):
  return getUtility(ISimpleSpsso)




# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Idp related views."""
from urllib import quote, unquote

from zope.schema import Choice, ASCIILine
from zope.formlib.form import Fields, action
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from Products.CMFCore.utils import getToolByName

from dm.zope.schema.z2.form import PageForm
from dm.zope.schema.widget import GenericTextWidget, make_hidden

from dm.zope.saml2.interfaces import _


class SelectIdp(PageForm):
  label = _(u"select_idp_title", u"Select your identity provider.")

  def __init__(self, context, request):

    def title(eid):
      """the title of entity *eid* (or *eid* itself)."""
      return getattr(context.get_entity(eid), "title", eid)

    self.form_fields = Fields(
      Choice(
        __name__=u"idp",
        title=_(u"identity_provider", u"Your identity provider"),
        vocabulary=SimpleVocabulary(
          tuple(SimpleTerm(eid, title(eid)) for eid in context.list_idps())
          ),
        default=context.default_idp,
        required=True,
        ),
      ASCIILine(
        __name__=u"came_from",
        default=request.get("came_from", ""),
        required=False,
        ),
      )
    super(SelectIdp, self).__init__(context, request)
    self.form_fields["came_from"].custom_widget = make_hidden(GenericTextWidget)

  @action(_(u"login", u"login"))
  def login(self, action, data):
    from dm.pdb import zpdb; zpdb.set_trace()
    c = self.context
    idp = data["idp"]
    c.set_idp_cookie(idp)
    return self.context.authn(idp, data["came_from"])




# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Authority and metadata."""
from datetime import datetime
from tempfile import NamedTemporaryFile

from persistent import Persistent
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from AccessControl import ClassSecurityInfo
#from Globals import InitializeClass

from zope.interface import Interface, implements
from zope.component import getUtility
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent

from OFS.SimpleItem import SimpleItem

import dm.xmlsec.binding as xmlsec

from dm.zope.schema.schema import SchemaConfigured, SchemaConfiguredEvolution
from dm.zope.schema.z2.constructor import SchemaConfiguredAddForm
from dm.saml2.metadata import EntityBase, EntityByUrl, \
     EntityMetadata, MetadataRepository, \
     role2element
from dm.saml2.binding import SoapBinding, HttpRedirectBinding, HttpPostBinding
from dm.saml2.pyxb.metadata import EndpointType, IndexedEndpointType
from dm.saml2 import signature

from interfaces import ISamlAuthority, \
     IIdpssoRole, ISpssoRole, IApRole, IAuthnRole, IPdpRole, \
     INameidFormatSupport

from permission import manage_saml
from util import ZodbSynchronized
from entity import ManageableEntityMixin, EntityManagerMixin


# Note: `EntityByUrl` lacks a way to access the context for signature
#  verification. This is a fundamental problem as we insist on
#  using only a priori known certificates for signature verification.
#  Such information is obviously missing for new entities.
#  Currently, we work around this by not verifying the metadata.
#  Later, we may enhance signature verification by allowing
#  subject certificates issued by a trusted CA.

# Note: we must inform the authority when metadata changes such that
#  it can update the signature verification context. We will
#  use events for this.
class IEntityMetadata(Interface):
  """marker interface."""

# Note: we do not derive from `Persistent` and do not use
#  a persistent list for internal purposes.
#  This is okay as long as we use a persistent storage (as we do).
#  Not using persistent classes here leads to less ZODB loads.
class EntityMetadata(EntityMetadata):
  """extend basic `EntityMetadata` by `ObjectModified` events."""
  implements(IEntityMetadata)

  def default_validity(self):
    return getUtility(ISamlAuthority).default_validity
  default_validity = property(default_validity, lambda self, value: None)

  def fetch_metadata(self, *args, **kw):
    super(EntityMetadata, self).fetch_metadata(*args, **kw)
    notify(ObjectModifiedEvent(self))

  def clear_metadata(self):
    super(EntityMetadata, self).clear_metadata()
    notify(ObjectModifiedEvent(self))

  def _get_metadata_sets(self, *args, **kw):
    rv = super(EntityMetadata, self)._get_metadata_sets(*args, **kw)
    if rv[0]: notify(ObjectModifiedEvent(self))
    return rv


class OwnEntity(ManageableEntityMixin, EntityBase):
  # Note: we cannot pass in the authority as this loses the acquition context
  #  Instead, we access it as utility. This assumes that it is global
  #  (an assumption used elsewhere, too).
  meta_type = "Saml own entity (cannot be deleted or changed)"

  manage_options = ManageableEntityMixin.manage_options[1:]

  protected = True # prevent this entity to be deleted

  # use the authorities entity id as our id
  def id(self):
    return getUtility(ISamlAuthority).entity_id
  id = property(id, lambda self, value: None)

  def get_metadata_document(self):
    return self._get_authority()._export_own_metadata()

  def _get_authority(self):
    return getUtility(ISamlAuthority)


class SamlAuthority(SchemaConfiguredEvolution, EntityManagerMixin,
                    SimpleItem, SchemaConfigured, MetadataRepository
                    ):
  """Zope2 implementation for a simple SAML authority.

  The implementation expects the instances to be persistent.

  In order to update internal data, the instance must be notified
  when metadata changes. Event handlers are set up to this effect
  which require that the instance is registered as a local utility
  in the context of the change.
  """
  meta_type = "Saml authority"

  implements(ISamlAuthority)

  INTERNAL_STORAGE_CLASS = PersistentMapping
  METADATA_STORAGE_CLASS = PersistentMapping
  METADATA_CLASS = EntityMetadata

  SC_SCHEMAS = (ISamlAuthority,)

  security = ClassSecurityInfo()

  security.declareObjectProtected(manage_saml)
  security.declareProtected("View", "metadata")

  manage_options = (
    {"label" : "Contents", "action" : "manage_main"},
    {"label" : "View", "action" : "@@view"},
    {"label" : "Edit", "action" : "@@edit"},
    {"label" : "Metadata", "action" : "metadata"},
    ) + SimpleItem.manage_options


  def __init__(self, **kw):
    SimpleItem.__init__(self)
    SchemaConfigured.__init__(self, **kw)
    MetadataRepository.__init__(self)
    # maps roles to paths to objects implementing the role
    self.roles = self.INTERNAL_STORAGE_CLASS()
    # add entity representing ourselves
    #  must be delayed until `ISamlAuthority` is registered
    #  likely, there are errors when `entityId` is changed
    # self.add_entity(OwnEntity())

  def register_role_implementor(self, implementor):
    # We allow *implementor* to implement more than a single role.
    # Role implementation is indicated by the implementation of the
    #   the respective interface (we may later add a *roles* parameter
    #   to restrict to a subset or allow implementor to restrict).
    md = globals(); roles = self.roles; path = implementor.getPhysicalPath()
    for role in role2element:
      i = md["I" + role.capitalize() + "Role"]
      if i.providedBy(implementor):
        if role in roles:
          raise ValueError("Role %s already registered" % role)
        roles[role] = path
    self._update()

  def unregister_role_implementor(self, implementor):
    roles = self.roles; path = implementor.getPhysicalPath()
    for r, ri in roles.items():
      if ri == path: del roles[r]
    self._update()

  def export_own_metadata(self):
    # return self._export_own_metadata()
    return self.metadata_by_id(self.entity_id).get_recent_metadata().toxml() # return the cached information

  def _update(self):
    """Something important has changed. Invalidate cached data."""
    self.metadata_by_id(self.entity_id).clear_metadata() # recreated on next use
    self._get_signature_cache().invalidate()

  def _export_own_metadata(self):
    """recompute our own metadata."""
    from dm.saml2.pyxb import metadata
    ed = metadata.EntityDescriptor(
      entityID=self.entity_id,
      validUntil=datetime.utcnow() + self.metadata_validity,
      )
    ld = metadata.__dict__
    for r, p in self.roles.items():
      if r == "ap": continue # for the moment
      i = self.unrestrictedTraverse(p)
      rd = ld[role2element[r]]()
      getattr(ed, rd.__class__.__name__[:-4]).append(rd)
      for c in (self.certificate, self.future_certificate):
        if c:
          c = _make_absolute(c)
          # build key_info
          from pyxb.bundles.wssplat.ds import KeyInfo, X509Data
          # this assumes the file to contain a (binary) X509v3 certificate
          cert = open(c, "rb").read()
          x509 = X509Data(); x509.X509Certificate = [cert]
          key_info = KeyInfo(); key_info.X509Data = [x509]
          rd.KeyDescriptor.append(
            metadata.KeyDescriptor(key_info, use="signing")
            )
      if hasattr(rd, "NameIDFormat"):
        nifs = INameidFormatSupport(i)
        rd.NameIDFormat = nifs.supported
      # add role specific information -- we do not yet support all roles
      getattr(self, "gen_metadata_" + r)(i, rd)
    return ed.toxml()

  def gen_metadata_sso(self, implementor, rd):
    pass
    # not yet supported
##    rd.ArtifactResolutionService.append(
##      IndexedEndpointType(
##        Binding=SoapBinding,
##        # may want to fix the protocol (to ensure "https" is used)
##        Location="%s/soap" % self._get_url(implementor),
##        index=0,
##        isDefault=True,
##        )
##      )
##    rd.SignleLogoutService.append(
##      EndpointType(
##        Binding=HttpPostBinding,
##        # may want to fix the protocol (to ensure "https" is used)
##        Location="%s/post" % self._get_url(implementor),
##        ))
##    rd.SignleLogoutService.append(
##      EndpointType(
##        Binding=HttpRedirectBinding,
##        # may want to fix the protocol (to ensure "https" is used)
##        Location="%s/redirect" % self._get_url(implementor),
##        ))
        
  def gen_metadata_idpsso(self, implementor, rd):
    self.gen_metadata_sso(implementor, rd)
    rd.SingleSignOnService.append(
      EndpointType(
        Binding=HttpRedirectBinding,
        # may want to fix the protocol (to ensure "https" is used)
        Location="%s/redirect" % self._get_url(implementor),
        ))
    rd.SingleSignOnService.append(
      EndpointType(
        Binding=HttpPostBinding,
        # may want to fix the protocol (to ensure "https" is used)
        Location="%s/post" % self._get_url(implementor),
        ))
    # should we support the artifact resolution binding?
    # might want to specify attributes here -- e.g. when we also implement attributes
    if IApRole.providedBy(implementor):
      implementor.add_attribute_metadata(rd)
        
  def gen_metadata_spsso(self, implementor, rd):
    self.gen_metadata_sso(implementor, rd)
    if implementor.wants_assertions_signed:
      rd.WantsAssertionsSigned = True
      # not safe enough (without signatures)
##    rd.AssertionConsumerService.append(
##      IndexedEndpointType(
##        Binding=HttpRedirectBinding,
##        # may want to fix the protocol (to ensure "https" is used)
##        Location="%s/redirect" % self._get_url(implementor),
##        index=0,
##        isDefault=False,
##        ))
    rd.AssertionConsumerService.append(
      IndexedEndpointType(
        Binding=HttpPostBinding,
        # may want to fix the protocol (to ensure "https" is used)
        Location="%s/post" % self._get_url(implementor),
        index=1,
        isDefault=True,
        ))
    # should we support artifact resolution?
    implementor.add_attribute_metadata(rd)

  def metadata(self, REQUEST):
    """Web access to our metadata."""
    R = REQUEST.response
    md = self.export_own_metadata()
    R.setHeader("Content-Type", "text/xml; charset=utf-8")
    R.setHeader("Cache-Control", "no-cache")
    R.setHeader("Pragma", "no-cache")
    R.setHeader("Expires", "0")
    return md

  def _get_url(self, obj):
    return self.base_url + "/" + obj.absolute_url(True)

  ## signature support
  _signature_cache = None

  def _get_signature_cache(self):
    sc = self._signature_cache
    if sc is None:
      sc = self._signature_cache = ZodbSynchronized()
    return sc

  def _get_signature_context(self, use):
    sc = self._get_signature_cache()
    vuse = "_v_" + use
    ctx = getattr(sc, vuse, None)
    if ctx is None:
      ctx = signature.SignatureContext()
      getattr(self, "_add_%s_keys" % use)(ctx)
      setattr(sc, vuse, ctx)
    return ctx

  def _add_sign_keys(self, ctx):
    ctx.add_key(xmlsec.Key.load(
      _make_absolute(self.private_key),
      xmlsec.KeyDataFormatPem,
      # if we have no private key password, pass some value
      #  to prevent an attempt to read it from the terminal
      #  We would like to use the *callback* parameter, but it
      #  does not work currently
      self.private_key_password and str(self.private_key_password) or "fail",
      ),
                self.entity_id
                )

  def _add_verify_keys(self, ctx):
    from dm.saml2.metadata import role2element
    for e in self.list_entities():
      seen = set()
      md = self.metadata_by_id(e.id)
      for r in role2element:
        for cert in md.get_role_certificates(r, "signing"):
          if cert in seen: continue
          seen.add(cert)
          ctx.add_key(
            xmlsec.Key.loadMemory(cert, xmlsec.KeyDataFormatCertDer, None),
            e.id
            )

  # override entity manager methods to update our signature context
  def _setOb(self, id, entity):
    super(SamlAuthority, self)._setOb(id, entity)
    self._get_signature_cache().invalidate()

  # no harm is done, not to invalidate on "_delOb".


# no longer necessary
#InitializeClass(SamlAuthority)


### automatic [un]registration on add/move/delete
def move_handler(o, e):
  # import will probably need to move for newer Zope
  from zope.app.component.hooks import getSite
  site = getSite()
  if site is None:
      raise ValueError("need a persistent active site")
  sm = site.getSiteManager()
  if e.oldParent:
    # prevent deletion when there are registered roles
    if o.roles and e.newParent is None:
      raise ValueError(
        "must first delete role implementers",
        ["/".join(p) for p in set(o.roles.values())]
        )
    # unregister
    sm.unregisterUtility(o, provided=ISamlAuthority)
  if e.newParent:
    # register
    # ensure this is a persistent registry
    if not hasattr(sm, "_p_jar"):
      raise ValueError("need a persistent active site")
    sm.registerUtility(o, provided=ISamlAuthority)
    # if `o` is new, we must add its `OwnEntity`
    if e.oldParent is None: # new
      o.add_entity(OwnEntity())


def own_metadata_changed(o, e):
  """subcriber to be informed whenever something changed that affects our own metadata."""
  getUtility(ISamlAuthority)._update()


def signature_context_changed(o, e):
  """subscriber to be informaed whenever the signature context may become invalid."""
  getUtility(ISamlAuthority)._get_signature_cache().invalidate()


### Customize add form
class AuthorityAddForm(SchemaConfiguredAddForm):
  def customize_fields(self):
    self.form_fields["base_url"].default = self.request["BASE_1"]


### Signature support
# To avoid passing through a signature/verification context, we
#  replace the default contexts in `signature` by objects
#  delegating to the saml authority

class _Delegator(object):
  def sign(self, *args, **kw):
    auth = self._get_authority()._get_signature_context("sign").sign(*args, **kw)

  def verify(self, *args, **kw):
    auth = self._get_authority()._get_signature_context("verify").verify(*args, **kw)

  def _get_authority(self):
    return getUtility(ISamlAuthority)

  # Note: we do not implement "add_key" as in our case, keys should
  #  be added only by the saml authority

signature.default_sign_context = signature.default_verify_context = _Delegator()


def _make_absolute(fn):
  """make *fn* an absolute filename, resolving relative names wrt `clienthome`."""
  from os.path import isabs, join
  if isabs(fn): return fn
  from App.config import getConfiguration
  return join(getConfiguration().clienthome, fn)



"""Entity management.

We assume that our entities are schema configured and implement `IEntity`.
"""
from urllib import quote, unquote

from zope.interface import Interface, implements
from zope.schema import URI, TextLine

from OFS.SimpleItem import SimpleItem
from OFS.ObjectManager import ObjectManager, IFAwareObjectManager
from AccessControl import ClassSecurityInfo

from dm.zope.schema.schema import SchemaConfigured

from dm.saml2.metadata import EntityByUrl

from dm.zope.saml2.interfaces import _, IEntity
from dm.zope.saml2.permission import manage_saml


class IManageableEntity(IEntity):
  specified_title = TextLine(
    title=_(u"specified_title_title", u"Title"),
    description=_(u"specified_title_description",
                  u"An explicitely specified title. If you do not specify a tilte, it defaults to the entity id."),
    required=False
    )



class ManageableEntityMixin(SchemaConfigured, SimpleItem):
  """Mixin class to provide `SchemaConfigured` based manageability.

  Note: We expect the entity class to be configurable by the
  schema only.

  The instances expect to be managed inside a `MetadataRepository`.
  """
  implements(IManageableEntity)

  manage_options = (
    {"label" : "Edit", "action" : "@@edit"},
    {"label" : "Metadata", "action" : "metadata"},
    {"label" : "Update", "action" : "@@update"},
    ) + SimpleItem.manage_options
    

  security = ClassSecurityInfo()
  security.declareObjectProtected(manage_saml)

  specified_title = ''

  def metadata(self, REQUEST):
    """Web access to metadata."""
    md = self.aq_inner.aq_parent.metadata_by_id(self.id).get_recent_metadata().toxml()
    R = REQUEST.response
    R.setHeader("Content-Type", "text/xml; charset=utf-8")
    R.setHeader("Cache-Control", "no-cache")
    R.setHeader("Pragma", "no-cache")
    R.setHeader("Expires", "0")
    return md
    
    
  # override `SimpleItem` methods such that our `id` attribute is used.
  def getId(self):
    # must return a quoted id as entity ids may contain characters
    #  forbidden in Zope ids
    return _quote(self.id)

  def _setId(self, id): self.id = id

  # implement `title` as `specified_title or id`.
  def _get_title(self): return self.specified_title or self.id
  def _set_title(self, title): self.specified_title = title
  title = property(_get_title, _set_title)


class IEntityByUrl(IManageableEntity):
  url = URI(
    title=_(u"entity_url_title", "Url"),
    description=_(
      u"entity_url_description",
      u"Url from which this entities' metadata can be fetched.",
      ),
    required=True,
    )



class EntityByUrl(ManageableEntityMixin, EntityByUrl):
  implements(IEntityByUrl)

  meta_type = "Saml2 entity defined by metadata providing url"

  SC_SCHEMAS = (IEntityByUrl,)


class EntityManagerMixin(IFAwareObjectManager, ObjectManager):
  """mixin class to manage managable entities.

  We assume to be part of a `MetadataRepository`.

  Note: we use quoted ids for the managed entities as the entity ids
    themselves can contain characters not allowed in Zope ids.
  """
  _product_interfaces = (IEntity,)

  def _setOb(self, id, entity):
    # *id* may or may not be quoted -- we ignore it altogether
    self.add_entity(entity)

  def _getOb(self, id, default=None):
    e = self.get_entity(_unquote(id), default)
    if e is not default: e = e.__of__(self) # acquisition wrap
    return e

  def _delOb(self, id):
    e = self.get_entity(_unquote(id))
    if e is None: return
    if getattr(e, "protected", False):
      raise TypeError(_(u"protected_entity_deletion",
                        u"Cannot delete a protected entity")
                      )
    self.del_entity(e.id)

  def __getitem__(self, k):
    e = self._getOb(k)
    if e is None: raise KeyError(k)
    return e

  def objectIds(self, spec=None):
    assert spec is None
    return [_quote(e.id) for e in self.list_entities()]

  def _checkId(self, id, allow_dup=0):
    qid = _quote(id)
    nid = super(EntityManagerMixin, self)._checkId(qid, allow_dup)
    return nid or qid

  def _setObject(self, id, object, *args, **kw):
    # the id may be quoted or unquoted -- we ignore it altogether.
    id = object.id
    return super(EntityManagerMixin, self)._setObject(id, object, *args, **kw)


def initialize(context):
  from dm.zope.schema.z2.constructor import add_form_factory, SchemaConfiguredZmiAddForm
  
  context.registerClass(
    EntityByUrl,
    constructors=(add_form_factory(EntityByUrl, form_class=SchemaConfiguredZmiAddForm),),
    permission=manage_saml,
    interfaces=(IEntity,),
    visibility=None,
    )


### auxiliaries
def _quote(id):
  # this does not yet allow all entity ids but should support the sensible ones
  return quote(id, "").replace("%", "$")

def _unquote(id):
  return unquote(id.replace("$", "%"))



from os.path import abspath, dirname, join
try:
  # try to use setuptools
  from setuptools import setup
  setupArgs = dict(
      include_package_data=True,
      install_requires=['dm.saml2>=2.0.1',
                        "dm.zope.schema>=2.0.1",
                        "setuptools", # to keep buildout happy
                        ],
      namespace_packages=['dm', 'dm.zope'],
      zip_safe=False,
      )
except ImportError:
  # use distutils
  from distutils import setup
  setupArgs = dict(
    )

cd = abspath(dirname(__file__))
pd = join(cd, 'dm', 'zope', 'saml2')

def pread(filename, base=pd): return open(join(base, filename)).read().rstrip()

setup(name='dm.zope.saml2',
      version=pread('VERSION.txt').split('\n')[0],
      description="Zope 2/Plone extension for SAML2 based Single Sign On: identity, attribute and service providers",
      long_description=pread('README.txt'),
      classifiers=[
#        'Development Status :: 3 - Alpha',
        "Development Status :: 4 - Beta",
#       'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: GNU General Public License (GPL)",
        'Framework :: Zope2',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities',
        ],
      author='Dieter Maurer',
      author_email='dieter@handshake.de',
      url='http://pypi.python.org/pypi/dm.zope.saml2',
      packages=['dm', 'dm.zope', 'dm.zope.saml2'],
      keywords='application development zope saml2 sso plone',
      license='ZPL',
      **setupArgs
      )



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""SAML2 support."""

def initialize(context):
  from authority import SamlAuthority
  from permission import manage_saml
  from dm.zope.schema.z2.constructor import add_form_factory, SchemaConfiguredZmiAddForm

  context.registerClass(
    SamlAuthority,
    constructors=(add_form_factory(SamlAuthority, form_class=SchemaConfiguredZmiAddForm),),
    permission=manage_saml,
    )

  from idpsso.idpsso import initialize; initialize(context)
  from spsso import initialize; initialize(context)
  from entity import initialize; initialize(context)



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
from Products.Five.browser import BrowserView
from Products.CMFCore.utils import getToolByName

from Products.statusmessages.interfaces import IStatusMessage

from dm.zope.saml2.interfaces import _


class Failure(BrowserView):
  def __call__(self):
    r = self.request
    IStatusMessage(r).addStatusMessage(_(u"SAML request failed"))
    r.response.redirect(getToolByName(self.context, "portal_url")())




# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Utilities."""
from zope.interface import alsoProvides
from zope.component import getUtility
from persistent import Persistent

def datetime_rfc822(dt):
  """format *dt* as an rfc822 date in GMT."""
  # this should be easier!
  from time import mktime
  from email.Utils import formatdate
  
  return formatdate(mktime(dt.timetuple()), usegmt=True)


def vocab_from_urns(urns):
  """turn sequence *urns* into a vocabulary.

  Each *urn* defines a term and the last `:` separated component the title.
  """
  from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

  return SimpleVocabulary(tuple(SimpleTerm(urn, title=urn.split(":")[-1])
                                for urn in urns
                                ))


# should propably be moved into a general utility package as it
#  could be usefull in a variety of situations
class UtilityRelay(object):
  """General utility implementing an object by a utility."""
  def __init__(self, interface, name=''):
    self.__interface = interface
    self.__name = name
    alsoProvides(self, interface)

  def getattr(self, k):
    if k not in self.__interface: raise AttributeError(k)
    return getattr(getUtility(self.__interface, self.__name), k)


# should propably be moved into a general utility package as it
#  could be usefull in a variety of situations
class ZodbSynchronized(Persistent):
  """auxiliary container for `_v_` variables synchronized via the ZODB.

  Used to maintain `_v_` caches consistent across several ZODB
  connections (in the same instance or across a range of them).
  """
  def invalidate(self):
    self._p_activate() # work around bug in some ZODB versions
    self._p_changed=True # ensure caches in other ZODB connections are flushed
    for k in self.__dict__.keys():
      if k.startswith("_v_"): delattr(self, k)



# Copyright (C) 2010-2012 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Exceptions"""

class SamlError(Exception):
  """A fatal SAML error. We are unable to send a request/response."""



# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""Generic role infrastructure."""
from datetime import datetime
from logging import getLogger
from os import environ

from zope.interface import implements
from zope.component import getUtility
from BTrees.OOBTree import OOBTree

from dm.saml2.pyxb.protocol import Response
from dm.saml2.pyxb.assertion import NameID, \
     SubjectType, Subject, SubjectConfirmation, \
     ConditionsCheckContext
from dm.saml2.binding import SoapBinding, HttpPostBinding, HttpRedirectBinding
from dm.saml2.util import normalize_nameid_format
from dm.saml2.binding.util import Store, UnmanagedError, RelayStateManager

from interfaces import ISamlAuthority, IHttpTransport, INameidFormatSupport
from exception import SamlError

logger = getLogger(__name__)

class Role(RelayStateManager):
  """Generic role infrastructure."""
  
  def __init__(self):
    RelayStateManager.__init__(self, Store(OOBTree()))


  def handle_Response(self, rsp, binding, relay_state):
    """handle a positive response."""
    rid, ok, fail = self._resolve_relay_state(relay_state)
    if rid != rsp.InResponseTo:
      raise SamlError("relay state and response to not match: %s" % rid)
    if binding == "redirect":
      raise SamlError("We do not accept responses via redirect (as we do not yet support signatures for redirect")
    context = ResponseContext(
      rsp, binding, self.REQUEST, self._get_authority(),
      )
    # check whether this is a success or failure
    if rsp.Status.StatusCode.Value != rsp.PREFIX + "Success":
      # this is a failure
      logger.error("failed SAML2 request: %s" % rsp.toxml())
      if fail is None: raise SystemError("unsolicited SAML failure response")
      return self.REQUEST.response.redirect(fail)
    # success response
    for ass in rsp.Assertion: self._process_assertion(ass, context)
    return self.REQUEST.response.redirect(ok)

  def _resolve_relay_state(self, relay_state):
    if relay_state:
      try: return self.retrieve(relay_state)
      except UnmanagedError:
        # an unsolicited response
        pass
    raise NotImplementedError("We do not support IDP initiated login")

  def _process_assertion(self, ass, context):
    # ensure we know the issuer -- an exception results, if not
    eid = ass.Issuer.value()
    auth = self._get_authority()
    auth.metadata_by_id(eid)
    if not ass.verified_signature():
      raise SamlError("assertion %s was not signed by the issuer" % ass.ID)
    if not ass.is_valid(context):
      raise SamlError("assertion %s is not valid" % ass.ID)
    if context.binding == "post":
      # check the `SubjectConfirmation` condition of the `Browser SSO profile`
      ok = False
      for sc in ass.Subject.SubjectConfirmation:
        scm = sc.Method
        if scm != "urn:oasis:names:tc:SAML:2.0:cm:bearer":
          # we do not understand the method -- hope, we find one we understand
          continue
        scd = sc.SubjectConfirmationData
        if scd is None: continue # not valid
        if scd.NotOnOrAfter is None or datetime.utcnow() >= scd.NotOnOrAfter:
          continue # not valid
        if scd.Recipient != context.rsp.Destination:
          continue # not valid
        if scd.InResponseTo != context.rsp.InResponseTo:
          continue # not valid
        if scd.Address and context.zrequest.getClientAddr() != scd.Address:
          continue # not valid
        ok = True
        break
      if not ok:
        raise SamlError("subject confirmation in assertion %s is invalid" % ass.ID)
    subject = ass.Subject.NameID
    for tag in ("Statement", "AuthnStatement", "AuthzDecisionStatement", "AttributeStatement"):
      for s in getattr(ass, tag):
        # process statement -- "AttributeError", if we do not support its type
        getattr(self, "_process_" + tag)(subject, s)



  def deliver_failure(self, target, req, fail, relay_state=None, msg=None):
    rsp = Response()
    rsp.set_status(("Responder", fail), msg)
    return self.deliver(target, rsp, req, relay_state)

  def deliver_success(self, target, req, assertions, relay_state=None):
    rsp = Response()
    rsp.set_success()
    rsp.Assertion = assertions
    return self.deliver(target, rsp, req, relay_state)

  def deliver(self, target, rsp, req, relay_state=None):
    """deliver *rsp* as response to *req* or *req* to *target*.

    If *rsp*, then a response is delivered - in response to *req* (if provided)
    otherwise, *req* is delivered.
    """
    auth = self._get_authority()
    eid = auth.entity_id
    target.resolve(auth, req, rsp)
    msg = rsp or req
    url = target.url
    msg.Destination = url
    msg.Issuer = NameID(eid)
    if target.sign_msg: msg.request_signature()
    if rsp and req: rsp.InResponseTo = req.ID
    if rsp:
      sign = target.sign_ass
      # postprocess assertions
      for ass in getattr(rsp, "Assertion", None):
        # add `Issuer` if it is not present
        if ass.Issuer is None: ass.Issuer = NameID(eid)
        # sign assertion if ours and requested for
        if sign and ass.Issuer.value() == eid: ass.request_signature()
    if logging_enabled:
      logger.info(
        "outgoing saml message to %s via binding %s with relay state %s:\n%s"
        % (msg.Destination,
           target.binding,
           relay_state,
           msg.toxml(root_only=True)
           )
        )
    handler = getattr(
      self,
      (rsp and self.response_handler or self.request_handler)[target.binding]
      )
    return handler(msg, relay_state)

  def _get_authority(self):
    return getUtility(ISamlAuthority)


  def make_assertion(self):
    """an empty assertion, issued by ourself."""
    from dm.saml2.pyxb.assertion import Assertion, NameID, Issuer
    return Assertion(NameID(self._get_authority().entity_id))


  # Note: SAML2 does not provide a good error status for the case
  #  that the subject does not match. We use `SUBJECT_MISMATCH` to let
  #  the status be easliy changed
  SUBJECT_MISMATCH = "UnknownPrinciple"

  def subject_from_member(self, member, target, req):
    """construct a subject description for *member*.

    The description is tailored for *target* and *req*.

    In case of a problem, error information (currently in the form of
    a string) it returned.
    """
    if member is None: return "AuthnFailed"
    auth = self._get_authority(); us = auth.entity_id
    teid = target.eid
    if teid is None: teid = target.eid = req.Issuer.value()
    subject = None
    # if *req* contains a subject, ensure it identifies *member*
    if req.Subject:
      # we only support `NameID` (and get an exception for something else)
      sn = req.Subject.NameID
      # ensure the subject specifies one of our subjects
      if sn.NameQualifier and sn.NameQualifier != us:
        return self.SUBJECT_MISMATCH
      subject = self._subject_from_member(member, sn.SPNameQualifier, sn.Format)
      if not isinstance(subject, SubjectType): return subject
      if sn.value() != subject.NameID.value(): return self.SUBJECT_MISMATCH
    # Note: we make assumptions below which might only hold for
    #  `AuthnRequest`.
    # determine the required name id policy
    unspecified = normalize_nameid_format("unspecified")
    format, sp, ok_create = None, teid, True
    if req.NameIDPolicy:
      nip = req.NameIDPolicy
      format, sp, ok_create = nip.Format, nip.SPNameQualifier, nip.AllowCreate
    if format in (None, unspecified):
      # find a format supported by us and the requester
      rd = target.get_role_descriptor(self._get_authority())
      supported = INameidFormatSupport(self).supported
      for nif in rd.NameIDFormat:
        if nif in supported: format = nif; break
      else:
        if not rd.NameIDFormat or unspecified in rd.NameIDFormat:
          # we can choose the format
          format = supported[0]
        else: return "InvalidNameIDPolicy"
    if sp is None: sp = teid
    return self._subject_from_member(member, sp, format, ok_create)

  def _subject_from_member(self, member, sp, format, ok_create=False):
    format = normalize_nameid_format(format)
    nifs = INameidFormatSupport(self)
    if format not in nifs.supported: return "InvalidNameIDPolicy"
    nid = NameID(NameQualifier=self._get_authority().entity_id, Format=format)
    subject = Subject(
      nid,
      )
    i = nifs.make_id(member, sp, ok_create, nid)
    if i is None: return "InvalidNameIDPolicy" # may not be adequate
    nid._resetContent(); nid.append(i)
    return subject


  # binding handlers
  request_handler = {
    HttpRedirectBinding : "http_redirect",
    HttpPostBinding : "http_post",
    SoapBinding: "soap_request",
    }

  response_handler = {
    HttpRedirectBinding : "http_redirect",
    HttpPostBinding : "http_post",
    SoapBinding: "soap_response",
    }

  def soap_request(self, msg, relay_state):
    from dm.saml2.binding.soap import http_post

    return http_post(getUtility(IHtmlTransport), msg.Destination, msg)

  def soap_response(self, msg, relay_state): return msg

  def http_redirect(self, msg, relay_state):
    from dm.saml2.binding.httpredirect import encode
    url = encode(msg.Destination, msg, relay_state)
    self.REQUEST.response.redirect(url)

  def http_post(self, msg, relay_state):
    from dm.saml2.binding.httppost import encode, as_controls
    return self.unrestrictedTraverse("@@saml_post_template")(
      action=msg.Destination,
      controls=as_controls(encode(msg, relay_state))
      )


class Target(object):
  """auxiliary class to capture destination information."""
  def __init__(self,
               role,     # the role this message is targeted at
               eid=None, # id of the SAML entity receiving the message
               binding=None, # the binding to use
               endpoint=None, # the message is targeted at
               endpoint_index=None, # the index associated with the endpoint
               url=None, # directly specifying the URL.
               sign_ass_attr=None, # role attribute calling for assertions to be signed
               sign_ass=None, # whether assertions should be signed
               sign_msg_attr=None, # role attribute calling for messages to be signed
               sign_msg=None, # whether the message should be signed
               ):
    for k,v in locals().items():
      if k != "self": setattr(self, k, v)

  def resolve(self, auth, req, rsp=None):
    """update the instance to have defined *url* and *binding*.

    Note: There is probably no need for *rsp*.
    """
    __traceback_info__ = self.__dict__
    eid = self.eid
    if eid is None and req is not None: eid = self.eid = req.Issuer.value()
    rd = self.get_role_descriptor(auth)
    if self.url:
      # the url has been directly specified.
      # SAML2 demands to verify that the url indeed belongs to the authority
      #  We try to ensure that be enforcing that the host in url
      #  is identical to the host part of one of the service endpoints
      #  used by this role.
      from dm.saml2.util import child_values
      from dm.saml2.pyxb.metadata import EndpointType
      from urlparse import urlsplit
      target_host = urlsplit(self.url)[1]
      ok = False
      for c in child_values(rd):
        if not isinstance(c, list): continue # all endpoints can appear multiple
        for ep in c:
          if not isinstance(ep, EndpointType): break # if the first value is not endpoint, none will be
          # should we consider `ResponseLocation` as well?
          if target_host == urlsplit(ep.Location)[1]: ok = True; break
        if ok: break
      if not ok:
        raise SamlError("`%s` does not belong to authority `%s`" % (self.url, eid))
      if self.binding is None:
 	# The request has specified a serice url but no
 	#  protocol binding. Try post (the only binding
 	#  we currently support for authentication responses)
        #  Alternatively, we may try the url to determine
        #  the binding from metadata
        self.binding = HttpPostBinding
    else:
      # `endpoint` must have been specified
      epd = getattr(rd, self.endpoint)
      # we may have more than a single endpoint
      if self.binding:
        # retrict by the requested binding
        epd = [d for d in epd if d.Binding == self.binding]
      epi = self.endpoint_index
      if epi is not None:
        # this requires the endpoints to be indexed endpoints
        epd = [d for d in epd if d.index == epi]
      if len(epd) > 1:
        # we still have more than a single candidate - select on `isDefault`
        epd = [d for d in epd if getattr(d, "isDefault", True)]
      if not epd:
        raise SamlError("no appropriate endpoint found: "
                        + str((eid, self.endpoint, self.binding, self.endpoint_index)))
      epd = epd[0]
      self.binding = epd.Binding
      self.url = rsp is not None and epd.ResponseLocation or epd.Location
    # determine whether we should sign the assertions
    # Note: SAML requires that assertions are signed in the
    #   HTTP-Post binding.
    if not self.sign_ass and self.sign_ass_attr:
      self.sign_ass = getattr(rd, self.sign_ass_attr)
    self.sign_ass = self.sign_ass or self.binding == HttpPostBinding
    # determine whether we should sign the message
    if not self.sign_msg and self.sign_msg_attr:
      self.sign_msg = getattr(rd, self.sign_msg_attr)


  def get_role_descriptor(self, auth):
    return auth.metadata_by_id(self.eid).get_role_descriptor(self.role)


class ResponseContext(ConditionsCheckContext):
  """Auxiliary class to encapsulate the context necessary for response verification."""
  def __init__(self, rsp, binding, zrequest, auth):
    self.rsp = rsp
    self.binding = binding # note: this is "redirect", "post" or "soap"
    self.zrequest = zrequest
    self.auth = auth

  def audience_id(self): return self.auth.entity_id



### automatic [un]registration on add/move/delete
def move_handler(o, e):
  auth = getUtility(ISamlAuthority)
  if e.oldParent:
    # unregister
    auth.unregister_role_implementor(o)
  if e.newParent:
    # register
    auth.register_role_implementor(o)


###### NameID format support

class NameidFormatSupport(object):
  """Default name id format support."""
  implements(INameidFormatSupport)

  def __init__(self, context): self.context = context


  def make_id(self, member, sp, ok_create, nid):
    return self._dispatcher[nid.Format](self, member, sp, ok_create, nid)

  def _unspecified(self, member, *unused): return member.getId()

  _dispatcher = {
    normalize_nameid_format("unspecified"):_unspecified,
    }

  supported = _dispatcher.keys()


######### Logging control
logging_enabled = bool(environ.get("SAML2_ENABLE_LOGGING", ''))
